import uuid
from os import listdir
from os.path import isfile, join
from biomage_programmatic_interface.sample_file import SampleFile
from biomage_programmatic_interface.utils import is_file_hidden

class Sample:

    def __init__(self, name):
        self.__name = name
        self.__uuid = str(uuid.uuid4())
        self.__sample_files = []
    
    def to_json(self):
        return {
            'name': self.__name,
            'sampleTechnology': '10x'
        }

    def name(self):
        return self.__name

    def uuid(self):
        return self.__uuid

    def experiment_id(self):
        return self.__experiment_id

    def get_sample_files(self):
        return self.__sample_files

    def add_sample_file(self, sample_file):
        self.__sample_files.append(sample_file)

    @staticmethod
    def __find_all_files_recursively(path):
        file_paths = listdir(path)
        ret = {}
        for file_path in file_paths:
            full_path = join(path, file_path)

            if is_file_hidden(full_path):
                continue

            if isfile(full_path):
                file = SampleFile(full_path)
                folder_name = file.folder()
    
                if ret.get(folder_name) == None:
                    ret[folder_name] = Sample(folder_name)
                ret[folder_name].add_sample_file(file)
                continue

            ret.update(Sample.__find_all_files_recursively(full_path))
        return ret

    @staticmethod
    def get_all_samples_from_path(path): 
        return Sample.__find_all_files_recursively(path).values()



from os.path import getsize
import gzip
import shutil
import requests
import uuid

class SampleFile:
    def __init__(self, path):
        self.__path = path
        self.__uuid = str(uuid.uuid4())

    @classmethod
    def from_sample_file(cls, sample_file):
        return SampleFile(sample_file.path())

    def path(self):
        return self.__path

    def name(self):
        return self.__path.split('/')[-1]

    def uuid(self):
        return self.__uuid

    def folder(self):
        return self.__path.split('/')[-2]

    def size(self):
        return getsize(self.__path)

    def type(self): 
        file_types = {
            'matrix': 'matrix10x',
            'barcodes': 'barcodes10x',
            'features': 'features10x',
            'genes': 'features10x',
        }
        print(self.name())
        for file_type_key in file_types.keys():
            if file_type_key in self.name():
                return file_types[file_type_key]
        return None

    def to_json(self):
        return {
            'sampleFileId': self.__uuid,
            'size': self.size(),
            "metadata": {},
        }

    def __is_compressed(self):
        with gzip.open(self.__path, 'r') as fh:
            try:
                # Try to read 1 byte of the file (Fails if not zipped)
                fh.read1()
                return True
            except:
                return False

    def __compress(self):
        with open(self.__path, 'rb') as f_in:
            compressed_url = self.__path + '.gz'
            with gzip.open(compressed_url, 'wb') as f_out:
                # Copying is done in chunks by default
                shutil.copyfileobj(f_in, f_out)
                self.__path = compressed_url

    def upload_to_S3(self, signed_url):
        if not self.__is_compressed():
            self.__compress()
        headers = { 'Content-type': 'application/octet-stream' }
        with open(self.__path, 'rb') as file:
            response = requests.put(signed_url, headers = headers, data = file.read())


import boto3
import requests
import datetime
import hashlib
import requests
import datetime
import re
from biomage_programmatic_interface.sample import Sample
from biomage_programmatic_interface.utils import load_json

class Connection:
    def __init__(self, username, password, instance='biomage'):
        self.__config = load_json('config.json')[instance]
        self.__try_authenticate(username, password)

    def __try_authenticate(self, username, password):
        client = boto3.client('cognito-idp', region_name='eu-west-1')

        try:
            resp = client.initiate_auth(
                ClientId=self.__config['ClientId'],
                AuthFlow='USER_PASSWORD_AUTH',
                AuthParameters = { 
                    "USERNAME": username,
                    "PASSWORD": password
                }
            )

            print('Authorization succesfull')
            self.__jwt = resp['AuthenticationResult']['IdToken']
        except:
            raise Exception("Incorrect username or password")

    def __fetch_api(self, url, json, method='POST'):
        methods = {
            'POST': requests.post,
            'PATCH': requests.patch
        }

        root_url = self.__config['URL']
        headers = {
            'Authorization': 'Bearer ' + self.__jwt,
            'Content-Type': 'application/json'
        }

        return methods[method](root_url + url, json=json, headers=headers)

    def create_experiment(self):
        created_at = datetime.datetime.now().isoformat()
        hashed_string = hashlib.md5(created_at.encode())
        experiment_id = hashed_string.hexdigest()

        experiment_data = {
            'id': experiment_id,
            'name': experiment_id,
            'description': ''
        }

        response = self.__fetch_api('v2/experiments/' + experiment_id, json=experiment_data)

        print('Experiment {} created!'.format(experiment_id))
        return experiment_id

    def __notify_upload(self, experiment_id, sample_id, sample_file_type):
        url = "v2/experiments/{}/samples/{}/sampleFiles/{}".format(experiment_id, sample_id, sample_file_type)
        json = {  
            "uploadStatus": "uploaded"
        }
        response = self.__fetch_api(url, json, 'PATCH')

    def __create_sample_file(self, experiment_id, sample_uuid, sample_file):     
        url = 'v2/experiments/{}/samples/{}/sampleFiles/{}'.format(experiment_id, sample_uuid, sample_file.type())
        response = self.__fetch_api(url, sample_file.to_json())
        return response.content

    def __create_and_upload_sample(self, experiment_id, sample):
        url = 'v2/experiments/{}/samples/{}'.format(experiment_id, sample.uuid())
        self.__fetch_api(url, sample.to_json())

        print('Created sample {} - {}'.format(sample.name(), sample.uuid()))

        for sample_file in sample.get_sample_files():
            s3url_raw = self.__create_sample_file(experiment_id, sample.uuid(), sample_file)
            print(sample_file)
            s3url = re.search(r"b\'\"(.*)\"\'", str(s3url_raw)).group(1)
            sample_file.upload_to_S3(s3url)
            self.__notify_upload(experiment_id, sample.uuid(), sample_file.type())
            print('Uploaded {} - {}...'.format(sample_file.name(), sample_file.uuid()))

    def upload_samples(self, experiment_id, samples_path):
        samples = Sample.get_all_samples_from_path(samples_path)
        for sample in samples:
            self.__create_and_upload_sample(experiment_id, sample)

        print('Project successfully created!')
